#######################################
#
# File: 19latentVariable.py
# Author: Michel Bierlaire, EPFL
# Date: Sun Nov 21 08:35:26 2010
#
#######################################

from biogeme import *
from headers import *
from distributions import *
from loglikelihood import *
from statistics import *

##
# Latent variable: income. Not directly observed. Respondents have
# only reported an income category, or have decided not to reveal it.
#
###
#Structural model: we assume, for the sake of the example, that income
#is a function of age and gender.
###

b_meanInc = Beta('b_meanInc',0.740017,0,10,0 )
b_gender = Beta('b_gender',0.375125,-10000,10000,0 )
b_age1 = Beta('b_age1',-1.72308,-10000,10000,0 )
b_age2 = Beta('b_age2',-0.201859,-10000,10000,0 )
b_sigma = Beta('b_sigma',0.01,-10,10,0 )


#Error term of the structural model
omega = RandomVariable('omega')
density = bioNormalPdf(omega)

# income is modeled in KCHF
income = b_meanInc + b_gender * MALE + b_age1 * (AGE == 1) + b_age2 * (AGE == 2) + b_sigma * omega

###
# Measurement equation: probability to reveal income in the questionnaire
####

r_reveal = Beta('r_reveal',0.97649,-1000,1000,0 )
r_income = Beta('r_income',0.823253,-1000,1000,0 )
r_male = Beta('r_male',1.26372,-1000,1000,0 )


utilReveal = r_reveal + r_income * income + r_male * MALE

probaReveal = exp(utilReveal) / (1 + exp(utilReveal))

###
# Measurement equation: ordered logit
# If the respondent decides to reveal the income, the response is discrete
###

# incomeCategory may take the following values:
#  1: under 50   (INCOME = 0 or INCOME = 1 in data file)
#  2: between 50 and 100 (INCOME = 2 in data file)
#  3: over 100 (INCOME = 3 in data file)
#  4: unkown (INCOME = 4 in data file)


incomeCategory = DefineVariable('incomeCategory',(INCOME <= 1) * 1 + (INCOME > 1) * INCOME)

scaleMeasurement = Beta('scaleMeasurement',3.91897,0.0001,1000,0 )

measurementEquation = {
    1: probaReveal * 1/(1+exp(scaleMeasurement * (income - 0.5))),
    2: probaReveal * (1/(1+exp(scaleMeasurement * (income - 1)))-1/(1+exp(scaleMeasurement * (income - 0.5)))),
    3: probaReveal * (1-1/(1+exp(scaleMeasurement * (income - 1)))),
    4: 1 - probaReveal }


#Parameters of the choice model to be estimated

ASC_CAR = Beta('ASC_CAR',-0.109243,-10,10,0 )
ASC_TRAIN = Beta('ASC_TRAIN',-0.689156,-10,10,0 )
ASC_SM = Beta('ASC_SM',0,-10,10,1)
B_TIME = Beta('B_TIME',-1.22523,-10,10,0 )
B_COST = Beta('B_COST',-0.82547,-10,10,0 )

# Utility functions

#If the person has a GA (season ticket) her incremental cost is actually 0 
#rather than the cost value gathered from the
# network data. 
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0. 
# A previous estimation with the unscaled data has generated
# parameters around -0.01 for both cost and time. Therefore, time and
# cost are multipled my 0.01.
# The following statements are designed to preprocess the data. It is
# like creating a new columns in the data file. This should be
# preferred to the statement like
# TRAIN_TT_SCALED = TRAIN_TT / 100.0
# which will cause the division to be reevaluated again and again,
# throuh the iterations. For models taking a long time to estimate, it
# may make a significant difference.
 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED', TRAIN_TT / 100.0)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED', TRAIN_COST / 100)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100)

# We use as a variable cost divided by income, which is the latent variable. 
V1 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED / income
V2 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED / income
V3 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED / income

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ))
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ))

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob = bioLogit(V,av,CHOICE)

# This is the joint probability of the stated choice and the stated income category
condLikelihoodOneObs = prob * Elem(measurementEquation,incomeCategory)

# It must be integrated over omega
likelihoodOneObs = Integrate(condLikelihoodOneObs * density,'omega')

# Defines an itertor on the data
rowIterator('obsIter') 

# Define the likelihood function for the estimation
BIOGEME_OBJECT.ESTIMATE = Sum(log(likelihoodOneObs),'obsIter')

# All observations verifying the following expression will not be
# considered for estimation
# The modeler here has developed the model only for work trips.  
# Observations such that the dependent variable CHOICE is 0 are also removed.
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) + ( CHOICE == 0 )) > 0

BIOGEME_OBJECT.EXCLUDE = exclude

# Statistics

nullLoglikelihood(av,'obsIter')
choiceSet = [1,2,3]
cteLoglikelihood(choiceSet,CHOICE,'obsIter')
availabilityStatistics(av,'obsIter')


BIOGEME_OBJECT.PARAMETERS['optimizationAlgorithm'] = "CFSQP"
BIOGEME_OBJECT.PARAMETERS['checkDerivatives'] = "1"
BIOGEME_OBJECT.PARAMETERS['numberOfThreads'] = "4"
