//-*-c++-*------------------------------------------------------------
//
// File name : bioArithBinaryPlus.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 15:25:26 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithBinaryPlus.h"
#include "bioArithConstant.h"
#include "bioExpressionRepository.h"

bioArithBinaryPlus::bioArithBinaryPlus(bioExpressionRepository* rep,
				       patULong par,
                                       patULong left, 
                                       patULong right,
				       patError*& err) 
  : bioArithBinaryExpression(rep, par,left,right,err) {

}

bioArithBinaryPlus::~bioArithBinaryPlus() {}

patString bioArithBinaryPlus::getOperatorName() const {
  return ("+") ;
}

patReal bioArithBinaryPlus::getValue(patBoolean prepareGradient,patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal left = leftChild->getValue(prepareGradient, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal right = rightChild->getValue(prepareGradient,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  return(left + right) ;

}

bioExpression* bioArithBinaryPlus::getDerivative(patULong aLiteralId,
						 patError*& err) const {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }
  
  bioExpression* leftResult = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  bioExpression* rightResult = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      // Left variable, right variable
      bioExpression* result = new bioArithBinaryPlus(theRepository,patBadId,leftResult->getId(),rightResult->getId(),err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      return result ;
    }
    else {
      // Left variable, right constant
      return leftResult ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      return rightResult ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = new bioArithConstant(theRepository,NULL,0) ;
      return result ;
    }
  }


}

bioArithBinaryPlus* bioArithBinaryPlus::getDeepCopy(bioExpressionRepository* rep,
						    patError*& err) const {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(rep,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithBinaryPlus* theNode = 
    new bioArithBinaryPlus(rep,patBadId,leftClone->getId(),rightClone->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
      
  return theNode ;
}

bioArithBinaryPlus* bioArithBinaryPlus::getShallowCopy(bioExpressionRepository* rep,
						    patError*& err) const {
  bioArithBinaryPlus* theNode = 
    new bioArithBinaryPlus(rep,patBadId,leftChild->getId(),rightChild->getId(),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
      
  return theNode ;
}

patString bioArithBinaryPlus::getExpression(patError*& err) const {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  
  patString l = leftChild->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }
  patString r = rightChild->getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }

  str << l << ' ' << getOperatorName() << ' ' << r  ;
  return patString(str.str()) ;
}

patString bioArithBinaryPlus::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {

    if (leftChild == NULL || rightChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
  
    patBoolean lSimple = leftChild->isSimple() ;
    patBoolean rSimple = rightChild->isSimple() ;

    patString vLeft ;
    if (lSimple) {
      vLeft = leftChild->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }
    }
    else {
	vLeft = leftChild->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      patString vRight ;
      if (rSimple) {
	vRight = rightChild->getSimpleCppCode(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      else {
	vRight = rightChild->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      stringstream str ;
      if (!lSimple) {
	str << vLeft << endl ;
      }
      if (!rSimple) {
	str << vRight << endl ;
      }
      str << getCppResultDeclaration(prefix) << "= (" ;
      if (lSimple) {
	str << vLeft ;
      }
      else {
	str << leftChild->getCppResult(prefix) ;
      }
      str << " ) + ( " ;
      if (rSimple) {
	str << vRight ;
      }
      else {
	str << rightChild->getCppResult(prefix) ;
      }
      str << ");" << endl ;
      str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
      str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
      str << "    }" << endl ;
      //      str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
      codeMustBeGenerated = patFALSE ;
      return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
  
}
  
patString bioArithBinaryPlus::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(" << vLeft << ")+(" << vRight << ")" ;
  return patString(str.str()) ;
}

patString bioArithBinaryPlus::getExpressionString() const {
  stringstream str ;
  str << '+' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}



bioFunctionAndDerivatives* bioArithBinaryPlus::getNumericalFunctionAndGradient(vector<patULong> literalIds,
							   patError*& err) {
  if (result.empty()) {
    result.resize(literalIds.size()) ;
  }
  bioFunctionAndDerivatives* l = leftChild->getNumericalFunctionAndGradient(literalIds,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  bioFunctionAndDerivatives* r = 
    rightChild->getNumericalFunctionAndGradient(literalIds,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  result.theFunction =  l->theFunction + r->theFunction ;
  for (patULong i = 0 ; i < literalIds.size() ; ++i) {
    result.theGradient[i] = l->theGradient[i] + r->theGradient[i] ;
  }
  return &result ;
}

