//-*-c++-*------------------------------------------------------------
//
// File name : bioArithNormalCdf.cc
// Author :    Michel Bierlaire
// Date :      Tue Jun  8 18:00:07 2010
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithNormalCdf.h"
#include "bioArithDivide.h"
#include "bioArithMult.h"
#include "bioArithConstant.h"
#include "bioArithExp.h"
#include "bioArithUnaryMinus.h"

bioArithNormalCdf::bioArithNormalCdf(bioExpressionRepository* rep,
				     patULong par,
				     patULong left,
				     patError*& err) 
  : bioArithUnaryExpression(rep,par,left,err)
{
}

bioArithNormalCdf::~bioArithNormalCdf() {}

patString bioArithNormalCdf::getOperatorName() const {
  return ("normalCdf") ;
}

patReal bioArithNormalCdf::getValue(patBoolean prepareGradient, 
				    patError*& err)   {
  if (child == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal childresult = child->getValue(prepareGradient,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal result = patNormalCdf::the()->compute(childresult,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  return result ;
}


bioExpression* bioArithNormalCdf::getDerivative(patULong aLiteralId, patError*& err) const {
  if (child == NULL) {
    err = new patErrNullPointer("patArithNode") ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  

  bioExpression* value1 = child->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  bioExpression* value2 = child->getDeepCopy(theRepository,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* deriv = child->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // value * value 

  bioExpression* y2 = new bioArithMult(theRepository,patBadId,value1->getId(),value2->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // (value * value) / 2

  bioExpression* two = new bioArithConstant(theRepository,patBadId,2.0) ;

  bioExpression* d = new bioArithDivide(theRepository,patBadId,y2->getId(),two->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // - (value * value) / 2

  bioExpression* minusd = new bioArithUnaryMinus(theRepository,patBadId,d->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // exp(- (value * value) / 2)

  bioExpression* expo = new bioArithExp(theRepository,patBadId,minusd->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  // oneDivSqrtTwoPi * exp(- (value * value) / 2)


  bioExpression*  oneDivSqrtTwoPi =  new bioArithConstant(theRepository,patBadId,1.0/sqrt(2*bioPi)) ;

  bioExpression* m = new bioArithMult(theRepository,patBadId,oneDivSqrtTwoPi->getId(),expo->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  bioExpression* result = new bioArithMult(theRepository,patBadId,m->getId(),deriv->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }

  return result ;
}


bioArithNormalCdf* bioArithNormalCdf::getDeepCopy(bioExpressionRepository* rep,
						  patError*& err) const {
  bioExpression* leftClone(NULL) ;
  if (child != NULL) {
    leftClone = child->getDeepCopy(rep,err) ;
  }

  bioArithNormalCdf* theNode = 
    new bioArithNormalCdf(rep,patBadId,leftClone->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return theNode ;
}

bioArithNormalCdf* bioArithNormalCdf::getShallowCopy(bioExpressionRepository* rep,
						  patError*& err) const {
  bioArithNormalCdf* theNode = 
    new bioArithNormalCdf(rep,patBadId,child->getId(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return theNode ;
}


patString bioArithNormalCdf::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = child->getCppCode(prefix, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }


  stringstream str ;
  str << vLeft << endl ;
  str << getCppResultDeclaration(prefix) << " ;" << endl ;
  str << "   " << getCppResult(prefix) << " = patNormalCdf::the()->compute(" << child->getCppResult(prefix) << ",err);" << endl ;
  str << "  if (err != NULL) {" << endl ;
  str << "    WARNING(err->describe()) ;" << endl ;
  str << "    return NULL;" << endl ;
  str << "  }" << endl ;
  
  //  str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
  codeMustBeGenerated = patFALSE ;
  
  return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patString bioArithNormalCdf::getSimpleCppCode(patError*& err) {
  if (child == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  patString vLeft = child->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "patNormalCdf()(" << vLeft << ")" ;
  return patString(str.str()) ;
}

patBoolean bioArithNormalCdf::isSimple() const {
  return patFALSE ;
}

patString bioArithNormalCdf::getExpressionString() const {
  stringstream str ;
  str << "$NC" ;
  if (child != NULL) {
    str << '{' << child->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;
}



bioFunctionAndDerivatives* bioArithNormalCdf::getNumericalFunctionAndGradient(vector<patULong> literalIds, patError*& err) {
  stringstream str ;
  str << "Not implemented for expression " << getOperatorName() ;
  err = new patErrMiscError(str.str()) ;
  WARNING(err->describe()) ;
  return NULL ;
  
}
