//-*-c++-*------------------------------------------------------------
//
// File name : bioRandomDraws.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Thu Jul 30 16:55:28 2009
//
//--------------------------------------------------------------------

#include <sstream>
#include "bioSample.h"
#include "patFileSize.h"
#include "bioLiteralRepository.h"
#include "bioRandomDraws.h"
#include "bioParameters.h"
#include "patErrMiscError.h"
#include "patErrOutOfRange.h"
#include "bioDrawIterator.h"
#include "patUnixUniform.h"
#include "patHalton.h"
#include "patHessTrain.h"
#include "patNormalWichura.h"
#include "patCenteredUniform.h"
#include "patZeroOneUniform.h"
#include "patZeroOneAntithetic.h"
#include "patAntithetic.h"

bioRandomDraws::bioRandomDraws() : draws(NULL), typeNames(4), rng(NULL),theNormalGenerator(NULL), theRectangularGenerator(NULL), theZeroOneGenerator(NULL), drawDim(0) {
  typeNames[NORMAL] = patString("Normal") ;
  typeNames[UNIFORM] = patString("Uniform") ;
  typeNames[UNIFORMSYM] = patString("UniformSymmetric") ;
  typeNames[UNDEFINED] = patString("Undefined") ;
}

bioRandomDraws::~bioRandomDraws() {
  if (draws != NULL) {
    for (patULong r = 0 ; r < drawDim ; ++r) {
      for (patULong rv = 0 ; rv < betaDim ; ++rv) {
	delete [] draws[r][rv] ;
      }
      delete [] draws[r] ;
    }
    delete [] draws ;
  }
}
bioRandomDraws* bioRandomDraws::the() {
  static bioRandomDraws* singleInstance = NULL ;

  if (singleInstance==NULL) {
    singleInstance = new bioRandomDraws ;
  }
  return singleInstance ;
}

patULong bioRandomDraws::addRandomVariable(patString name, 
					   bioRandomDraws::bioDrawsType type,
					   patString index,
					   patError*& err) {
  //  DEBUG_MESSAGE("Add variable " << name << " of type " << typeNames[type]) ;
  map<patString,patULong>::iterator found = idFromName.find(name) ;
  if (found != idFromName.end()) {
    if (types[found->second] != type) {
      stringstream str;
      str << "Variable " << name << " has two different types: " 
	  << typeNames[types[found->second]] << " and " << typeNames[type] ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patBadId;
    }
    return found->second;
  }
  patULong id = names.size() ;
  names.push_back(name) ;
  indexNames.push_back(index) ;
  patULong indexId = bioLiteralRepository::the()->getLiteralId(index, err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patBadId ;
  }
  indexIds.push_back(indexId) ;
  listOfIds.push_back(vector<patULong>()) ;
  types.push_back(type) ;
  idFromName[name] = id ;
  return id ;
}


void bioRandomDraws::generateDraws(patError*& err) {

  if (!hasRandomVariables()) {
    return ;
  }
  drawDim = bioParameters::the()->getValueInt("NbrOfDraws",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  int dump = bioParameters::the()->getValueInt("dumpDrawsOnFile",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  // Largest ID
  patULong largestId(0) ;
  for (patULong rv = 0 ; rv < types.size() ; ++rv) {
    for (vector<patULong>::iterator ids = listOfIds[rv].begin() ;
	 ids != listOfIds[rv].end();
	 ++ids) {
      if (*ids > largestId) {
	largestId = *ids ;
      }
    }
    DEBUG_MESSAGE("Variable " << names[rv] << " Largest ID=" << largestId) ;
  }

  betaDim = types.size() ;
  idDim = largestId+1 ;
  draws = new patReal**[drawDim] ;
  for (patULong r = 0 ; r < drawDim ; ++r) {
    draws[r] = new patReal*[betaDim] ;
    for (patULong rv = 0 ; rv < betaDim ; ++rv) {
      draws[r][rv] = new patReal[idDim] ;
    }
  }
  DETAILED_MESSAGE("Generate " << drawDim << " draws for " << betaDim << " random parameters") ;
  DETAILED_MESSAGE("Size reserved in memory: " << patFileSize(betaDim * drawDim * idDim * sizeof(patReal)) << " = " << idDim << " * " << betaDim << " * " << drawDim << " * " << patFileSize(sizeof(patReal)) ) ;
  DETAILED_MESSAGE("Sizeof returns: " << sizeof(draws)) ;
  
  // Init random number generators

  patString rvType = bioParameters::the()->getValueString("RandomDistribution",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  patULong seed = bioParameters::the()->getValueInt("Seed",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  if (rvType == "HALTON") {
    DETAILED_MESSAGE("Prepare Halton draws for " << types.size() << " random parameters") ;
    patULong maxPrime = bioParameters::the()->getValueInt("maxPrimeNumbers",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
    rng = new patHalton(types.size(),
			maxPrime,
			drawDim,
			err) ;
    if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
    }
  }
  else if (rvType == "MLHS") {
    
    patUnixUniform* arng = new patUnixUniform(seed) ;
    rng = new patHessTrain(drawDim,arng) ;
  }
  else if (rvType == "PSEUDO") {
    rng = new patUnixUniform(seed) ;
  }
  else {
    err = new patErrMiscError("Unknown value of the parameter RandomDistribution. Valid entries are \"HALTON\", \"PSEUDO\" and \"MLHS\"") ;
    WARNING(err->describe()) ;
    return ;
  }
  
  patNormalWichura* theNormal = new patNormalWichura() ;
  theNormal->setUniform(rng) ;

  patBoolean antithetic = (bioParameters::the()->getValueInt("antitheticDraws",err) != 0);
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  if (antithetic) {
    theNormalGenerator = new patAntithetic(theNormal,drawDim,largestId+1,dump,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  else {
    theNormalGenerator = theNormal ;
  }
  
  if (antithetic) {
    patZeroOneAntithetic* theZeroOne = new patZeroOneAntithetic(drawDim,largestId+1,dump,err) ;
    theZeroOne->setUniform(rng) ;
    theZeroOneGenerator = theZeroOne ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  else {
    patZeroOneUniform* theZeroOne = new patZeroOneUniform(dump) ; 
    theZeroOne->setUniform(rng) ;
    theZeroOneGenerator = theZeroOne ;
  }

  patCenteredUniform* theRect = new patCenteredUniform(dump) ;
  theRect->setUniform(rng) ;
  if (antithetic) {
    theRectangularGenerator = new patAntithetic(theRect,
						drawDim,
						largestId+1,
						dump,
						err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  else {
    theRectangularGenerator = theRect ;
  }

  
  for (patULong rv = 0 ; rv < types.size() ; ++rv) {
    for (vector<patULong>::iterator ids = listOfIds[rv].begin() ;
	 ids != listOfIds[rv].end();
	 ++ids) {
      for (patULong r = 0 ; r < drawDim ; ++r) {
	switch (types[rv]) {
	case NORMAL:
	  //	  DEBUG_MESSAGE("Normal draws[" <<r << "][" << rv << "][" << *ids << "]") ;
	  draws[r][rv][*ids] = theNormalGenerator->getNextValue(err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  break ;
	case UNIFORM:
	  //	  	  DEBUG_MESSAGE("Uniform draws[" <<r << "][" << rv << "][" << *ids << "]") ;
	  draws[r][rv][*ids] = theZeroOneGenerator->getNextValue(err) ;
	  break ;
	case UNIFORMSYM:
	  //	  	  DEBUG_MESSAGE("Uniform symmetric draws[" <<r << "][" << rv << "][" << *ids << "]") ;
	  draws[r][rv][*ids] = theRectangularGenerator->getNextValue(err) ;
	  break ;
	case UNDEFINED:
	default:
	  err = new patErrMiscError("Undefined type for random variable") ;
	  WARNING(err->describe()) ;
	  return ;
	}
      }
    }
  }


  if (dump != 0) {
    ofstream df("draws.lis") ;
    for (patULong rv = 0 ; rv < types.size() ; ++rv) {
      for (vector<patULong>::iterator ids = listOfIds[rv].begin() ;
	   ids != listOfIds[rv].end();
	   ++ids) {
	df << names[rv] << "[" << *ids << "]" << '\t' ;
      }
    }
    df << endl ;
  
  
    for (patULong r = 0 ; r < drawDim ; ++r) {
      for (patULong rv = 0 ; rv < types.size() ; ++rv) {
	for (vector<patULong>::iterator ids = listOfIds[rv].begin() ;
	     ids != listOfIds[rv].end();
	     ++ids) {
	  df << draws[r][rv][*ids] << '\t' ;
	}
	df << endl ;
      }
    }
    df.close() ;
  }
  
}

//vector< vector<patReal> >* bioRandomDraws::getRow(patULong rowId, patError*& err) {
patReal** bioRandomDraws::getRow(patULong rowId, patError*& err) {
  if (rowId >= drawDim) {
    err = new patErrOutOfRange<patULong>(rowId,0,drawDim-1) ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  return draws[rowId] ;
}

patULong bioRandomDraws::getColId(patString name, patError*& err) {
  map<patString,patULong>::iterator found = idFromName.find(name) ;
  if (found != idFromName.end()) {
    return found->second ;
  }
  else {
    return patBadId ;
  }
  
}

bioDrawIterator* bioRandomDraws::createIterator() {
  //  bioDrawIterator* ptr = new bioDrawIterator(&draws) ;
  bioDrawIterator* ptr = new bioDrawIterator(draws,drawDim) ;
  return ptr ;
}

patBoolean bioRandomDraws::hasRandomVariables() const {
  return !names.empty() ;
}

void bioRandomDraws::populateListOfIds(bioSample* aSample, patError*& err) {
  for (patULong i = 0 ; i < indexNames.size() ; ++i) {
    vector<patULong> theList = aSample->getListOfIds(indexNames[i],err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    listOfIds[i] = theList ;
  }
}

patULong bioRandomDraws::nbrOfDraws() const {
  //  return draws.size() ;
  return drawDim ;
}

ostream& operator<<(ostream& str, const bioRandomDraws& x) {
  
  str << "Number of draws:   " << x.drawDim << endl ;
  if (x.drawDim > 0) {
    str << "Number of variables: " << x.betaDim << endl ;
    if (x.betaDim > 0) {
      str << "Number of obs/ind:     " << x.idDim << endl ;
      
      for (patULong obs = 0 ; obs < x.idDim ; ++obs) {
	str << "Observation " << obs << endl ;
	str << "*****************" << endl ;
	for (patULong var = 0 ; var < x.betaDim ;++var) {
	  str << "Draws for " << x.names[var] << ": " ;
	  for (patULong d = 0 ; d <  x.drawDim ; ++d) {
	    str << x.draws[d][var][obs] << " " ;
	  }
	  str << endl ;
	}
	str << endl ;
      }
    }
  }
  return str ;


  // str << "Number of draws:   " << x.draws.size() << endl ;
  // if (x.draws.size() > 0) {
  //   str << "Number of variables: " << x.draws[0].size() << endl ;
  //   if (x.draws[0].size() > 0) {
  //     str << "Number of obs/ind:     " << x.draws[0][0].size() << endl ;
      
  //     for (patULong obs = 0 ; obs < x.draws[0][0].size() ; ++obs) {
  // 	str << "Observation " << obs << endl ;
  // 	str << "*****************" << endl ;
  // 	for (patULong var = 0 ; var < x.draws[0].size() ;++var) {
  // 	  str << "Draws for " << x.names[var] << ": " ;
  // 	  for (patULong d = 0 ; d <  x.draws.size() ; ++d) {
  // 	    str << x.draws[d][var][obs] << " " ;
  // 	  }
  // 	  str << endl ;
  // 	}
  // 	str << endl ;
  //     }
  //   }
  // }
  // return str ;
}
